#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


class RPCWalletError(Exception):
    """Raised when remote wallet returned an error"""


class InvalidWalletError(Exception):
    """Raised when wrong wallet class passed"""


class EmptyWalletError(Exception):
    """Raised when wrong wallet class passed"""


class SChainNotFoundException(Exception):
    """Raised when requested sChain is not found"""


class IncompatibleAbiError(Exception):
    """Raised when required contract is not found in the ABI file"""
    def __init__(self, key):
        message = f'Required field was not found in the ABI file: {key}'
        super().__init__(message)


class InvalidNodeIdError(Exception):
    """Raised when wrong node id passed"""
    def __init__(self, node_id):
        message = f'Node with ID = {node_id} doesn\'t exist!'
        super().__init__(message)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.contracts.base_contract import BaseContract, transaction_method


class ConstantsHolder(BaseContract):
    @transaction_method
    def set_periods(self, new_reward_period, new_delta_period):
        return self.contract.functions.setPeriods(
            new_reward_period,
            new_delta_period
        )

    def get_reward_period(self):
        return self.contract.functions.rewardPeriod().call()

    def get_delta_period(self):
        return self.contract.functions.deltaPeriod().call()

    @transaction_method
    def set_check_time(self, new_check_time):
        return self.contract.functions.setCheckTime(new_check_time)

    def get_check_time(self):
        return self.contract.functions.checkTime().call()

    @transaction_method
    def set_latency(self, new_allowable_latency):
        return self.contract.functions.setLatency(new_allowable_latency)

    def get_latency(self):
        return self.contract.functions.allowableLatency().call()

    def get_first_delegation_month(self):
        return self.contract.functions.firstDelegationsMonth().call()

    def msr(self) -> int:
        """Minimum staking requirement to create a node.

        :returns: MSR (in wei)
        :rtype: int
        """
        return self.contract.functions.msr().call()

    @transaction_method
    def _set_msr(self, new_msr: int) -> None:
        """For internal usage only"""
        return self.contract.functions.setMSR(new_msr)

    def get_launch_timestamp(self) -> int:
        return self.contract.functions.launchTimestamp().call()

    @transaction_method
    def set_launch_timestamp(self, launch_timestamp: int):
        return self.contract.functions.setLaunchTimestamp(launch_timestamp)

    @transaction_method
    def set_rotation_delay(self, rotation_delay: int) -> None:
        """For internal usage only"""
        return self.contract.functions.setRotationDelay(rotation_delay)

    def get_rotation_delay(self) -> int:
        return self.contract.functions.rotationDelay().call()

    def get_dkg_timeout(self) -> int:
        return self.contract.functions.complaintTimelimit().call()

    @transaction_method
    def set_complaint_timelimit(self, complaint_timelimit: int):
        return self.contract.functions.setComplaintTimelimit(complaint_timelimit)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging

from sgx import SgxClient
from web3 import Web3

from skale.utils.web3_utils import get_eth_nonce
from skale.wallets.common import BaseWallet


logger = logging.getLogger(__name__)


class SgxWallet(BaseWallet):
    def __init__(self, sgx_endpoint, web3, key_name=None, path_to_cert=None):
        self.sgx_client = SgxClient(sgx_endpoint, path_to_cert=path_to_cert)
        self._web3 = web3
        if key_name is None:
            self._key_name, self._address, self._public_key = self._generate()
        else:
            self._key_name = key_name
            self._address, self._public_key = self._get_account(key_name)

    def sign(self, tx_dict):
        if tx_dict.get('nonce') is None:
            tx_dict['nonce'] = get_eth_nonce(self._web3, self._address)
        return self.sgx_client.sign(tx_dict, self.key_name)

    def sign_and_send(self, tx_dict) -> str:
        signed_tx = self.sign(tx_dict)
        return self._web3.eth.sendRawTransaction(signed_tx.rawTransaction).hex()

    def sign_hash(self, unsigned_hash: str):
        if unsigned_hash.startswith('0x'):
            unsigned_hash = unsigned_hash[2:]

        body = bytes.fromhex(unsigned_hash)
        header = b'\x19Ethereum Signed Message:\n32'
        normalized_hash = header + body
        hash_to_sign = Web3.keccak(hexstr='0x' + normalized_hash.hex())
        chain_id = None
        return self.sgx_client.sign_hash(hash_to_sign, self._key_name, chain_id)

    @property
    def address(self):
        return self._address

    @property
    def public_key(self):
        return self._public_key

    @property
    def key_name(self):
        return self._key_name

    def _generate(self):
        key = self.sgx_client.generate_key()
        return key.name, key.address, key.public_key

    def _get_account(self, key_name):
        account = self.sgx_client.get_account(key_name)
        return account.address, account.public_key



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" Monitors.sol functions """

import socket
from web3 import Web3
from skale.contracts.base_contract import BaseContract
from skale.utils.helper import format_fields

FIELDS = ['id', 'rep_date', 'ip']


class Monitors(BaseContract):
    def _get_checked_array_raw(self, node_id):
        node_id_bytes = Web3.solidityKeccak(['uint256'], [node_id])
        return self.contract.functions.getCheckedArray(node_id_bytes).call()

    @format_fields(FIELDS, flist=True)
    def get_checked_array_struct(self, node_id):
        return self._get_checked_array_raw(node_id)

    def get_checked_array(self, node_id):
        checked_array = self.get_checked_array_struct(node_id)
        for node in checked_array:
            node['ip'] = socket.inet_ntoa(node['ip'])
        return checked_array

    def get_last_bounty_block(self, node_index):
        return self.contract.functions.getLastBountyBlock(node_index).call()

    def get_last_received_verdict_block(self, node_index):
        return self.contract.functions.getLastReceivedVerdictBlock(
            node_index).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SchainsInternal.sol functions """

import functools
from skale.contracts.base_contract import BaseContract


class SChainsInternal(BaseContract):
    """Wrapper for some of the SchainsInternal.sol functions"""

    @property
    @functools.lru_cache()
    def schains(self):
        return self.skale.schains

    def get_raw(self, name):
        return self.contract.functions.schains(name).call()

    def get_all_schains_ids(self):
        return self.contract.functions.getSchains().call()

    def get_schains_number(self):
        return self.contract.functions.numberOfSchains().call()

    def get_schain_list_size(self, account):
        return self.contract.functions.getSchainListSize(account).call(
            {'from': account})

    def get_schain_id_by_index_for_owner(self, account, index):
        return self.contract.functions.schainIndexes(account, index).call()

    def get_node_ids_for_schain(self, name):
        id_ = self.schains.name_to_id(name)
        return self.contract.functions.getNodesInGroup(id_).call()

    def get_schain_ids_for_node(self, node_id):
        return self.contract.functions.getSchainIdsForNode(node_id).call()

    def is_schain_exist(self, name):
        id_ = self.schains.name_to_id(name)
        return self.contract.functions.isSchainExist(id_).call()

    def get_active_schain_ids_for_node(self, node_id):
        return self.contract.functions.getActiveSchains(node_id).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum


class SkaledPorts(Enum):
    PROPOSAL = 0
    CATCHUP = 1
    WS_JSON = 2
    HTTP_JSON = 3
    BINARY_CONSENSUS = 4
    ZMQ_BROADCAST = 5
    MTA = 6
    WSS_JSON = 7
    HTTPS_JSON = 8
    INFO_HTTP_JSON = 9



# flake8: noqa

from skale.contracts.contract_manager import ContractManager
from skale.contracts.base_contract import BaseContract, transaction_method

from skale.contracts.ima.lock_and_data import LockAndData
